@@ -18,13 +18,19 @@ from install import build
 from install.build import PLATFORM_LINUX
 from install.build import PLATFORM_WIN32
 
+try:
+    # This is to avoid getting numpy imported inside other modules and
+    # overwritting setuptools compilers (#5476)
+    import numpy.distutils  # NOQA
+except Exception:
+    pass
 
 # Cython requirements (minimum version and versions known to be broken).
 # Note: this must be in sync with setup_requires defined in setup.py.
 required_cython_version = pkg_resources.parse_version('0.29.22')
 ignore_cython_versions = [
 ]
-use_hip = bool(int(os.environ.get('CUPY_INSTALL_USE_HIP', '0')))
+use_hip = build.use_hip
 
 
 # The value of the key 'file' is a list that contains extension names
@@ -47,33 +53,33 @@ cuda_files = [
     'cupy_backends.cuda.libs.nvrtc',
     'cupy_backends.cuda.libs.profiler',
     'cupy_backends.cuda.stream',
-    'cupy.core._accelerator',
-    'cupy.core._carray',
-    'cupy.core._cub_reduction',
-    'cupy.core._dtype',
-    'cupy.core._fusion_kernel',
-    'cupy.core._fusion_thread_local',
-    'cupy.core._fusion_trace',
-    'cupy.core._fusion_variable',
-    'cupy.core._kernel',
-    'cupy.core._memory_range',
-    'cupy.core._optimize_config',
-    'cupy.core._reduction',
-    'cupy.core._routines_binary',
-    'cupy.core._routines_indexing',
-    'cupy.core._routines_linalg',
-    'cupy.core._routines_logic',
-    'cupy.core._routines_manipulation',
-    'cupy.core._routines_math',
-    'cupy.core._routines_sorting',
-    'cupy.core._routines_statistics',
-    'cupy.core._scalar',
-    'cupy.core.core',
-    'cupy.core.flags',
-    'cupy.core.internal',
-    'cupy.core.fusion',
-    'cupy.core.new_fusion',
-    'cupy.core.raw',
+    'cupy._core._accelerator',
+    'cupy._core._carray',
+    'cupy._core._cub_reduction',
+    'cupy._core._dtype',
+    'cupy._core._fusion_kernel',
+    'cupy._core._fusion_thread_local',
+    'cupy._core._fusion_trace',
+    'cupy._core._fusion_variable',
+    'cupy._core._kernel',
+    'cupy._core._memory_range',
+    'cupy._core._optimize_config',
+    'cupy._core._reduction',
+    'cupy._core._routines_binary',
+    'cupy._core._routines_indexing',
+    'cupy._core._routines_linalg',
+    'cupy._core._routines_logic',
+    'cupy._core._routines_manipulation',
+    'cupy._core._routines_math',
+    'cupy._core._routines_sorting',
+    'cupy._core._routines_statistics',
+    'cupy._core._scalar',
+    'cupy._core.core',
+    'cupy._core.flags',
+    'cupy._core.internal',
+    'cupy._core.fusion',
+    'cupy._core.new_fusion',
+    'cupy._core.raw',
     'cupy.cuda.common',
     'cupy.cuda.cufft',
     'cupy.cuda.device',
@@ -84,25 +90,43 @@ cuda_files = [
     'cupy.cuda.stream',
     'cupy.cuda.texture',
     'cupy.fft._cache',
+    'cupy.fft._callback',
     'cupy.lib._polynomial',
     'cupy._util'
 ]
 
 if use_hip:
+    # We handle nvtx (and likely any other future support) here, because
+    # the HIP stubs (hip/cupy_*.h) would cause many symbols
+    # to leak into all these modules even if unused. It's easier for all of
+    # them to link to the same set of shared libraries.
     MODULES.append({
         # TODO(leofang): call this "rocm" or "hip" to avoid confusion?
         'name': 'cuda',
-        'file': cuda_files,
+        'required': True,
+        'file': cuda_files + [
+            'cupy_backends.cuda.libs.nvtx',
+            'cupy_backends.cuda.libs.cusolver',
+            'cupy.cusolver',
+        ],
         'include': [
             'hip/hip_runtime_api.h',
             'hip/hiprtc.h',
             'hipblas.h',
             'hiprand/hiprand.h',
+            'hipfft.h',
+            'roctx.h',
+            'rocsolver.h',
         ],
         'libraries': [
             'amdhip64',  # was hiprtc and hip_hcc before ROCm 3.8.0
             'hipblas',
+            ('hipfft', lambda hip_version: hip_version >= 401),
             'hiprand',
+            'rocfft',
+            'roctx64',
+            'rocblas',
+            'rocsolver',
         ],
         'check_method': build.check_hip_version,
         'version_method': build.get_hip_version,
@@ -110,6 +134,7 @@ if use_hip:
 else:
     MODULES.append({
         'name': 'cuda',
+        'required': True,
         'file': cuda_files,
         'include': [
             'cublas_v2.h',
@@ -134,20 +159,13 @@ else:
         'version_method': build.get_cuda_version,
     })
 
-if use_hip:
-    MODULES.append({
-        'name': 'cusolver',
-        'file': [
-            'cupy_backends.cuda.libs.cusolver',
-        ],
-        'include': [],
-        'libraries': [],
-    })
-else:
+if not use_hip:
     MODULES.append({
         'name': 'cusolver',
+        'required': True,
         'file': [
             'cupy_backends.cuda.libs.cusolver',
+            'cupy.cusolver',
         ],
         'include': [
             'cusolverDn.h',
@@ -178,7 +196,7 @@ if not use_hip:
     MODULES.append({
         'name': 'nccl',
         'file': [
-            'cupy.cuda.nccl',
+            'cupy_backends.cuda.libs.nccl',
         ],
         'include': [
             'nccl.h',
@@ -193,7 +211,7 @@ if not use_hip:
     MODULES.append({
         'name': 'nvtx',
         'file': [
-            'cupy.cuda.nvtx',
+            'cupy_backends.cuda.libs.nvtx',
         ],
         'include': [
             'nvToolsExt.h',
@@ -223,6 +241,7 @@ if not use_hip:
 
     MODULES.append({
         'name': 'cub',
+        'required': True,
         'file': [
             ('cupy.cuda.cub', ['cupy/cuda/cupy_cub.cu']),
         ],
@@ -235,9 +254,77 @@ if not use_hip:
         'check_method': build.check_cub_version,
         'version_method': build.get_cub_version,
     })
+
+    MODULES.append({
+        'name': 'jitify',
+        'required': True,
+        'file': [
+            'cupy.cuda.jitify',
+        ],
+        'include': [
+            'cuda.h',
+            'cuda_runtime.h',
+            'nvrtc.h',
+        ],
+        'libraries': [
+            'cuda',
+            'cudart',
+            'nvrtc',
+        ],
+        'check_method': build.check_jitify_version,
+        'version_method': build.get_jitify_version,
+    })
+
+    MODULES.append({
+        'name': 'random',
+        'required': True,
+        'file': [
+            'cupy.random._bit_generator',
+            ('cupy.random._generator_api',
+             ['cupy/random/cupy_distributions.cu']),
+        ],
+        'include': [
+        ],
+        'libraries': [
+            'cudart',
+            'curand',
+        ],
+    })
+
+    MODULES.append({
+        'name': 'cusparselt',
+        'file': [
+            'cupy_backends.cuda.libs.cusparselt',
+        ],
+        'include': [
+            'cusparseLt.h',
+        ],
+        'libraries': [
+            'cusparseLt',
+        ],
+        'check_method': build.check_cusparselt_version,
+        'version_method': build.get_cusparselt_version,
+    })
+
+    MODULES.append({
+        'name': 'cugraph',
+        'file': [
+            'cupy_backends.cuda.libs.cugraph',
+        ],
+        'include': [
+            'cugraph/raft/error.hpp',  # dummy
+        ],
+        'libraries': [
+            'cugraph',
+        ],
+        'check_method': build.check_cugraph_version,
+        'version_method': build.get_cugraph_version,
+    })
+
 else:
     MODULES.append({
         'name': 'cub',
+        'required': True,
         'file': [
             ('cupy.cuda.cub', ['cupy/cuda/cupy_cub.cu']),
         ],
@@ -251,10 +338,26 @@ else:
         'version_method': build.get_cub_version,
     })
 
+    MODULES.append({
+        'name': 'nccl',
+        'file': [
+            'cupy_backends.cuda.libs.nccl',
+        ],
+        'include': [
+            'rccl.h',
+        ],
+        'libraries': [
+            'rccl',
+        ],
+        'check_method': build.check_nccl_version,
+        'version_method': build.get_nccl_version,
+    })
+
 if bool(int(os.environ.get('CUPY_SETUP_ENABLE_THRUST', 1))):
     if use_hip:
         MODULES.append({
             'name': 'thrust',
+            'required': True,
             'file': [
                 ('cupy.cuda.thrust', ['cupy/cuda/cupy_thrust.cu']),
             ],
@@ -268,6 +371,7 @@ if bool(int(os.environ.get('CUPY_SETUP_ENABLE_THRUST', 1))):
     else:
         MODULES.append({
             'name': 'thrust',
+            'required': True,
             'file': [
                 ('cupy.cuda.thrust', ['cupy/cuda/cupy_thrust.cu']),
             ],
@@ -285,8 +389,9 @@ if bool(int(os.environ.get('CUPY_SETUP_ENABLE_THRUST', 1))):
 
 MODULES.append({
     'name': 'dlpack',
+    'required': True,
     'file': [
-        'cupy.core.dlpack',
+        'cupy._core.dlpack',
     ],
     'include': [
         'cupy/dlpack/dlpack.h',
@@ -325,6 +430,10 @@ def module_extension_sources(file, use_cython, no_cuda):
     return [pyx] + others
 
 
+def get_required_modules():
+    return [m['name'] for m in MODULES if m.get('required', False)]
+
+
 def check_readthedocs_environment():
     return os.environ.get('READTHEDOCS', None) == 'True'
 
@@ -350,6 +459,20 @@ def check_library(compiler, includes=(), libraries=(),
     return True
 
 
+def canonicalize_hip_libraries(hip_version, libraries):
+    def ensure_tuple(x):
+        return x if isinstance(x, tuple) else (x, None)
+    new_libraries = []
+    for library in libraries:
+        lib_name, pred = ensure_tuple(library)
+        if pred is None:
+            new_libraries.append(lib_name)
+        elif pred(hip_version):
+            new_libraries.append(lib_name)
+    libraries.clear()
+    libraries.extend(new_libraries)
+
+
 def preconfigure_modules(compiler, settings):
     """Returns a list of modules buildable in given environment and settings.
 
@@ -408,6 +531,31 @@ def preconfigure_modules(compiler, settings):
                     settings['library_dirs'].append(lib_path)
                     break
 
+        if module['name'] == 'cugraph':
+            cugraph_path = os.environ.get('CUGRAPH_PATH', '')
+            for i in 'include', 'include/cugraph':
+                inc_path = os.path.join(cugraph_path, i)
+                if os.path.exists(inc_path):
+                    settings['include_dirs'].append(inc_path)
+            lib_path = os.path.join(cugraph_path, 'lib')
+            if os.path.exists(lib_path):
+                settings['library_dirs'].append(lib_path)
+
+        # In ROCm 4.1 and later, we need to use the independent version of
+        # hipfft as well as rocfft. We configure the lists of include
+        # directories and libraries to link here depending on ROCm version
+        # before the configuration process following.
+        if use_hip and module['name'] == 'cuda':
+            if module['check_method'](compiler, settings):
+                hip_version = module['version_method']()
+                if hip_version >= 401:
+                    rocm_path = build.get_rocm_path()
+                    inc_path = os.path.join(rocm_path, 'hipfft', 'include')
+                    settings['include_dirs'].insert(0, inc_path)
+                    lib_path = os.path.join(rocm_path, 'hipfft', 'lib')
+                    settings['library_dirs'].insert(0, lib_path)
+                canonicalize_hip_libraries(hip_version, module['libraries'])
+
         print('')
         print('-------- Configuring Module: {} --------'.format(
             module['name']))
@@ -433,7 +581,7 @@ def preconfigure_modules(compiler, settings):
             # Fail on per-library condition check (version requirements etc.)
             installed = True
             errmsg = ['The library is installed but not supported.']
-        elif (module['name'] in ('thrust', 'cub')
+        elif (module['name'] in ('thrust', 'cub', 'random')
                 and (nvcc_path is None and hipcc_path is None)):
             installed = True
             cmd = 'nvcc' if not use_hip else 'hipcc'
@@ -459,6 +607,10 @@ def preconfigure_modules(compiler, settings):
             if module['name'] == 'cuda':
                 break
 
+    # Get a list of the CC of the devices connected to this node
+    if not use_hip:
+        build.check_compute_capabilities(compiler, settings)
+
     if len(ret) != len(MODULES):
         if 'cuda' in ret:
             lines = [
@@ -526,6 +678,9 @@ def make_extensions(options, compiler, use_cython):
         settings['define_macros'].append(('CUPY_NO_CUDA', '1'))
     if use_hip:
         settings['define_macros'].append(('CUPY_USE_HIP', '1'))
+        # introduced since ROCm 4.2.0
+        settings['define_macros'].append(('__HIP_PLATFORM_AMD__', '1'))
+        # deprecated since ROCm 4.2.0
         settings['define_macros'].append(('__HIP_PLATFORM_HCC__', '1'))
 
     available_modules = []
@@ -533,7 +688,8 @@ def make_extensions(options, compiler, use_cython):
         available_modules = [m['name'] for m in MODULES]
     else:
         available_modules, settings = preconfigure_modules(compiler, settings)
-        if 'cuda' not in available_modules:
+        required_modules = get_required_modules()
+        if not (set(required_modules) <= set(available_modules)):
             raise Exception('Your CUDA environment is invalid. '
                             'Please check above error log.')
 
@@ -550,8 +706,8 @@ def make_extensions(options, compiler, use_cython):
         link_args = s.setdefault('extra_link_args', [])
 
         if module['name'] == 'cusolver':
-            compile_args = s.setdefault('extra_compile_args', [])
-            link_args = s.setdefault('extra_link_args', [])
+            # cupy_backends/cupy_lapack.h has C++ template code
+            compile_args.append('--std=c++11')
             # openmp is required for cusolver
             if use_hip:
                 pass
@@ -561,14 +717,27 @@ def make_extensions(options, compiler, use_cython):
             elif compiler.compiler_type == 'msvc':
                 compile_args.append('/openmp')
 
+        if module['name'] == 'random':
+            if compiler.compiler_type == 'msvc':
+                compile_args.append('-D_USE_MATH_DEFINES')
+
+        if module['name'] == 'jitify':
+            # this fixes RTD (no_cuda) builds...
+            compile_args.append('--std=c++11')
+            # if any change is made to the Jitify header, we force recompiling
+            s['depends'] = ['./cupy/_core/include/cupy/jitify/jitify.hpp']
+
         if module['name'] == 'dlpack':
             # if any change is made to the DLPack header, we force recompiling
-            s['depends'] = ['./cupy/core/include/cupy/dlpack/dlpack.h']
+            s['depends'] = ['./cupy/_core/include/cupy/dlpack/dlpack.h']
 
         for f in module['file']:
             s_file = copy.deepcopy(s)
             name = module_extension_name(f)
 
+            if name.endswith('fft._callback') and not PLATFORM_LINUX:
+                continue
+
             rpath = []
             if not options['no_rpath']:
                 # Add library directories (e.g., `/usr/local/cuda/lib64`) to
@@ -588,7 +757,7 @@ def make_extensions(options, compiler, use_cython):
 
             if not PLATFORM_WIN32 and not PLATFORM_LINUX:
                 assert False, "macOS is no longer supported"
-            if (PLATFORM_LINUX and s_file['library_dirs']):
+            if (PLATFORM_LINUX and len(rpath) != 0):
                 ldflag = '-Wl,'
                 if PLATFORM_LINUX:
                     ldflag += '--disable-new-dtags,'
@@ -754,12 +923,17 @@ def cythonize(extensions, arg_options):
     if compile_time_env is None:
         compile_time_env = {}
         cythonize_options['compile_time_env'] = compile_time_env
-    compile_time_env['use_hip'] = arg_options['use_hip']
+    compile_time_env['CUPY_CUFFT_STATIC'] = False
     compile_time_env['cython_version'] = str(cython_version)
-    if use_hip or arg_options['no_cuda']:
+    if arg_options['no_cuda']:  # on RTD
         compile_time_env['CUDA_VERSION'] = 0
-    else:
+        compile_time_env['HIP_VERSION'] = 0
+    elif use_hip:  # on ROCm/HIP
+        compile_time_env['CUDA_VERSION'] = 0
+        compile_time_env['HIP_VERSION'] = build.get_hip_version()
+    else:  # on CUDA
         compile_time_env['CUDA_VERSION'] = build.get_cuda_version()
+        compile_time_env['HIP_VERSION'] = 0
 
     return Cython.Build.cythonize(
         extensions, verbose=True, language_level=3,
@@ -798,69 +972,84 @@ def _nvcc_gencode_options(cuda_version):
         return []
 
     envcfg = os.getenv('CUPY_NVCC_GENERATE_CODE', None)
-    if envcfg:
+    if envcfg is not None and envcfg != 'current':
         return ['--generate-code={}'.format(arch)
                 for arch in envcfg.split(';') if len(arch) > 0]
-
-    # The arch_list specifies virtual architectures, such as 'compute_61', and
-    # real architectures, such as 'sm_61', for which the CUDA input files are
-    # to be compiled.
-    #
-    # The syntax of an entry of the list is
-    #
-    #     entry ::= virtual_arch | (virtual_arch, real_arch)
-    #
-    # where virtual_arch is a string which means a virtual architecture and
-    # real_arch is a string which means a real architecture.
-    #
-    # If a virtual architecture is supplied, NVCC generates a PTX code for the
-    # virtual architecture. If a pair of a virtual architecture and a real
-    # architecture is supplied, NVCC generates a PTX code for the virtual
-    # architecture as well as a cubin code for the real architecture.
-    #
-    # For example, making NVCC generate a PTX code for 'compute_60' virtual
-    # architecture, the arch_list has an entry of 'compute_60'.
-    #
-    #     arch_list = ['compute_60']
-    #
-    # For another, making NVCC generate a PTX code for 'compute_61' virtual
-    # architecture and a cubin code for 'sm_61' real architecture, the
-    # arch_list has an entry of ('compute_61', 'sm_61').
-    #
-    #     arch_list = [('compute_61', 'sm_61')]
-    #
-    # See the documentation of each CUDA version for the list of supported
-    # architectures:
-    #
-    #   https://docs.nvidia.com/cuda/cuda-compiler-driver-nvcc/index.html#options-for-steering-gpu-code-generation
-
-    if cuda_version >= 11000:
-        arch_list = ['compute_35',
-                     'compute_50',
-                     ('compute_60', 'sm_60'),
-                     ('compute_61', 'sm_61'),
-                     ('compute_70', 'sm_70'),
-                     ('compute_75', 'sm_75'),
-                     ('compute_80', 'sm_80'),
-                     'compute_80']
-    elif cuda_version >= 10000:
-        arch_list = ['compute_30',
-                     'compute_50',
-                     ('compute_60', 'sm_60'),
-                     ('compute_61', 'sm_61'),
-                     ('compute_70', 'sm_70'),
-                     ('compute_75', 'sm_75'),
-                     'compute_70']
-    elif cuda_version >= 9000:
-        arch_list = ['compute_30',
-                     'compute_50',
-                     ('compute_60', 'sm_60'),
-                     ('compute_61', 'sm_61'),
-                     ('compute_70', 'sm_70'),
-                     'compute_70']
+    if envcfg == 'current' and build.get_compute_capabilities() is not None:
+        ccs = build.get_compute_capabilities()
+        arch_list = [
+            f'compute_{cc}' if cc < 60 else (f'compute_{cc}', f'sm_{cc}')
+            for cc in ccs]
     else:
-        # This should not happen.
-        assert False
+        # The arch_list specifies virtual architectures, such as 'compute_61',
+        # and real architectures, such as 'sm_61', for which the CUDA
+        # input files are to be compiled.
+        #
+        # The syntax of an entry of the list is
+        #
+        #     entry ::= virtual_arch | (virtual_arch, real_arch)
+        #
+        # where virtual_arch is a string which means a virtual architecture and
+        # real_arch is a string which means a real architecture.
+        #
+        # If a virtual architecture is supplied, NVCC generates a PTX code
+        # the virtual architecture. If a pair of a virtual architecture and a
+        # real architecture is supplied, NVCC generates a PTX code for the
+        # virtual architecture as well as a cubin code for the real one.
+        #
+        # For example, making NVCC generate a PTX code for 'compute_60' virtual
+        # architecture, the arch_list has an entry of 'compute_60'.
+        #
+        #     arch_list = ['compute_60']
+        #
+        # For another, making NVCC generate a PTX code for 'compute_61' virtual
+        # architecture and a cubin code for 'sm_61' real architecture, the
+        # arch_list has an entry of ('compute_61', 'sm_61').
+        #
+        #     arch_list = [('compute_61', 'sm_61')]
+        #
+        # See the documentation of each CUDA version for the list of supported
+        # architectures:
+        #
+        #   https://docs.nvidia.com/cuda/cuda-compiler-driver-nvcc/index.html#options-for-steering-gpu-code-generation
+
+        if cuda_version >= 11010:
+            arch_list = ['compute_35',
+                         'compute_50',
+                         ('compute_60', 'sm_60'),
+                         ('compute_61', 'sm_61'),
+                         ('compute_70', 'sm_70'),
+                         ('compute_75', 'sm_75'),
+                         ('compute_80', 'sm_80'),
+                         ('compute_86', 'sm_86'),
+                         'compute_86']
+        elif cuda_version >= 11000:
+            arch_list = ['compute_35',
+                         'compute_50',
+                         ('compute_60', 'sm_60'),
+                         ('compute_61', 'sm_61'),
+                         ('compute_70', 'sm_70'),
+                         ('compute_75', 'sm_75'),
+                         ('compute_80', 'sm_80'),
+                         'compute_80']
+        elif cuda_version >= 10000:
+            arch_list = ['compute_30',
+                         'compute_50',
+                         ('compute_60', 'sm_60'),
+                         ('compute_61', 'sm_61'),
+                         ('compute_70', 'sm_70'),
+                         ('compute_75', 'sm_75'),
+                         'compute_70']
+        elif cuda_version >= 9020:
+            arch_list = ['compute_30',
+                         'compute_50',
+                         ('compute_60', 'sm_60'),
+                         ('compute_61', 'sm_61'),
+                         ('compute_70', 'sm_70'),
+                         'compute_70']
+        else:
+            # This should not happen.
+            assert False
 
     options = []
     for arch in arch_list:
@@ -888,22 +1077,28 @@ class _UnixCCompiler(unixccompiler.UnixCCompiler):
         if use_hip:
             return self._compile_unix_hipcc(
                 obj, src, ext, cc_args, extra_postargs, pp_opts)
+        else:
+            return self._compile_unix_nvcc(
+                obj, src, ext, cc_args, extra_postargs, pp_opts)
 
+    def _compile_unix_nvcc(self,
+                           obj, src, ext, cc_args, extra_postargs, pp_opts):
         # For CUDA C source files, compile them with NVCC.
+        nvcc_path = build.get_nvcc_path()
+        base_opts = build.get_compiler_base_options(nvcc_path)
+        compiler_so = nvcc_path
+
+        cuda_version = build.get_cuda_version()
+        postargs = _nvcc_gencode_options(cuda_version) + [
+            '-O2', '--compiler-options="-fPIC"']
+        if cuda_version >= 11020:
+            postargs += ['--std=c++14']
+            num_threads = int(os.environ.get('CUPY_NUM_NVCC_THREADS', '2'))
+            postargs += [f'-t{num_threads}']
+        else:
+            postargs += ['--std=c++11']
+        print('NVCC options:', postargs)
         try:
-            nvcc_path = build.get_nvcc_path()
-            base_opts = build.get_compiler_base_options(nvcc_path)
-            compiler_so = nvcc_path
-
-            cuda_version = build.get_cuda_version()
-            postargs = _nvcc_gencode_options(cuda_version) + [
-                '-O2', '--compiler-options="-fPIC"']
-            if cuda_version >= 11020:
-                postargs += ['--std=c++14']
-            else:
-                postargs += ['--std=c++11']
-            print('NVCC options:', postargs)
-
             self.spawn(compiler_so + base_opts + cc_args + [src, '-o', obj] +
                        postargs)
         except errors.DistutilsExecError as e:
@@ -912,14 +1107,18 @@ class _UnixCCompiler(unixccompiler.UnixCCompiler):
     def _compile_unix_hipcc(self,
                             obj, src, ext, cc_args, extra_postargs, pp_opts):
         # For CUDA C source files, compile them with HIPCC.
+        rocm_path = build.get_hipcc_path()
+        base_opts = build.get_compiler_base_options(rocm_path)
+        compiler_so = rocm_path
+
+        hip_version = build.get_hip_version()
+        postargs = ['-O2', '-fPIC', '--include', 'hip_runtime.h']
+        if hip_version >= 402:
+            postargs += ['--std=c++14']
+        else:
+            postargs += ['--std=c++11']
+        print('HIPCC options:', postargs)
         try:
-            rocm_path = build.get_hipcc_path()
-            base_opts = build.get_compiler_base_options(rocm_path)
-            compiler_so = rocm_path
-
-            postargs = ['-O2', '-fPIC', '--include', 'hip_runtime.h']
-            print('HIPCC options:', postargs)
-
             self.spawn(compiler_so + base_opts + cc_args + [src, '-o', obj] +
                        postargs)
         except errors.DistutilsExecError as e:
@@ -929,13 +1128,13 @@ class _UnixCCompiler(unixccompiler.UnixCCompiler):
         use_hipcc = False
         if use_hip:
             for i in objects:
-                if 'cupy_thrust.o' in i:
+                if any(obj in i for obj in ('cupy_thrust.o', 'cupy_cub.o')):
                     use_hipcc = True
         if use_hipcc:
             _compiler_cxx = self.compiler_cxx
             try:
-                rcom_path = build.get_hipcc_path()
-                self.set_executable('compiler_cxx', rcom_path)
+                rocm_path = build.get_hipcc_path()
+                self.set_executable('compiler_cxx', rocm_path)
 
                 return unixccompiler.UnixCCompiler.link(
                     self, target_desc, objects, output_filename, *args)
@@ -969,10 +1168,12 @@ class _MSVCCompiler(msvccompiler.MSVCCompiler):
             # to build CuPy because some Python versions were built using it.
             # REF: https://wiki.python.org/moin/WindowsCompilers
             postargs += ['-allow-unsupported-compiler']
-        postargs += ['-Xcompiler', '/MD']
+        postargs += ['-Xcompiler', '/MD', '-D_USE_MATH_DEFINES']
         # This is to compile thrust with MSVC2015
         if cuda_version >= 11020:
             postargs += ['--std=c++14']
+            num_threads = int(os.environ.get('CUPY_NUM_NVCC_THREADS', '2'))
+            postargs += [f'-t{num_threads}']
         print('NVCC options:', postargs)
 
         for obj in objects:
@@ -1040,7 +1241,7 @@ class custom_build_ext(build_ext.build_ext):
         build_ext.build_ext.run(self)
 
     def build_extensions(self):
-        num_jobs = int(os.environ.get('CUPY_NUM_BUILD_JOBS', '1'))
+        num_jobs = int(os.environ.get('CUPY_NUM_BUILD_JOBS', '4'))
         if num_jobs > 1:
             self.parallel = num_jobs
         super().build_extensions()
