@@ -1,5 +1,6 @@
 import contextlib
 import distutils.util
+import importlib
 import os
 import re
 import shutil
@@ -7,14 +8,27 @@ import subprocess
 import sys
 import tempfile
 
-from install import utils
+
+# This is done because pip is installed in no editable mode in
+# windows CI so the kernel cache gets always a deterministic path.
+def _from_install_import(name):
+    install_module_path = os.path.join(os.path.dirname(__file__))
+    original_sys_path = sys.path.copy()
+    try:
+        sys.path.append(install_module_path)
+        return importlib.import_module(name)
+    finally:
+        sys.path = original_sys_path
+
+
+utils = _from_install_import('utils')
 
 
 PLATFORM_LINUX = sys.platform.startswith('linux')
 PLATFORM_WIN32 = sys.platform.startswith('win32')
 
-minimum_cuda_version = 9000
-minimum_cudnn_version = 7000
+minimum_cuda_version = 9020
+minimum_cudnn_version = 7600
 
 minimum_hip_version = 305  # for ROCm 3.5.0+
 
@@ -22,6 +36,8 @@ _cuda_path = 'NOT_INITIALIZED'
 _rocm_path = 'NOT_INITIALIZED'
 _compiler_base_options = None
 
+use_hip = bool(int(os.environ.get('CUPY_INSTALL_USE_HIP', '0')))
+
 
 # Using tempfile.TemporaryDirectory would cause an error during cleanup
 # due to a bug: https://bugs.python.org/issue26660
@@ -147,9 +163,10 @@ def get_compiler_setting(use_hip):
     if rocm_path:
         include_dirs.append(os.path.join(rocm_path, 'include'))
         include_dirs.append(os.path.join(rocm_path, 'include', 'hip'))
-        include_dirs.append(os.path.join(rocm_path, 'rocrand', 'include'))
+        include_dirs.append(os.path.join(rocm_path, 'include', 'rocrand'))
+        include_dirs.append(os.path.join(rocm_path, 'include', 'hiprand'))
+        include_dirs.append(os.path.join(rocm_path, 'include', 'roctracer'))
         library_dirs.append(os.path.join(rocm_path, 'lib'))
-        library_dirs.append(os.path.join(rocm_path, 'rocrand', 'lib'))
 
     if use_hip:
         extra_compile_args.append('-std=c++11')
@@ -170,7 +187,9 @@ def get_compiler_setting(use_hip):
 
     # for <cupy/complex.cuh>
     cupy_header = os.path.join(os.path.dirname(os.path.realpath(__file__)),
-                               '../cupy/core/include')
+                               '../cupy/_core/include')
+    global _jitify_path
+    _jitify_path = os.path.join(cupy_header, 'cupy/jitify')
     if cuda_path:
         cuda_cub_path = os.path.join(cuda_path, 'include', 'cub')
         if not os.path.exists(cuda_cub_path):
@@ -279,6 +298,11 @@ _nccl_version = None
 _cutensor_version = None
 _cub_path = None
 _cub_version = None
+_jitify_path = None
+_jitify_version = None
+_compute_capabilities = None
+_cusparselt_version = None
+_cugraph_version = None
 
 
 def check_cuda_version(compiler, settings):
@@ -302,7 +326,7 @@ def check_cuda_version(compiler, settings):
     if _cuda_version < minimum_cuda_version:
         utils.print_warning(
             'CUDA version is too old: %d' % _cuda_version,
-            'CUDA 9.0 or newer is required')
+            'CUDA 9.2 or newer is required')
         return False
 
     return True
@@ -357,6 +381,43 @@ def get_hip_version(formatted=False):
     return _hip_version
 
 
+def check_compute_capabilities(compiler, settings):
+    """Return compute capabilities of the installed devices."""
+    global _compute_capabilities
+    try:
+        src = '''
+        #include <cuda_runtime_api.h>
+        #include <stdio.h>
+        #define CHECK_CUDART(x) { if ((x) != cudaSuccess) return 1; }
+
+        int main() {
+          int device_count;
+          CHECK_CUDART(cudaGetDeviceCount(&device_count));
+          for (int i = 0; i < device_count; i++) {
+              cudaDeviceProp prop;
+              CHECK_CUDART(cudaGetDeviceProperties(&prop, i));
+              printf("%d%d ", prop.major, prop.minor);
+          }
+          return 0;
+        }
+        '''
+        out = build_and_run(
+            compiler, src,
+            include_dirs=settings['include_dirs'],
+            libraries=('cudart',),
+            library_dirs=settings['library_dirs'])
+        _compute_capabilities = set([int(o) for o in out.split()])
+    except Exception as e:
+        utils.print_warning('Cannot check compute capability\n{0}'.format(e))
+        return False
+
+    return True
+
+
+def get_compute_capabilities(formatted=False):
+    return _compute_capabilities
+
+
 def check_thrust_version(compiler, settings):
     global _thrust_version
 
@@ -434,22 +495,29 @@ def check_nccl_version(compiler, settings):
 
     # NCCL 1.x does not provide version information.
     try:
-        out = build_and_run(compiler, '''
-        #include <nccl.h>
-        #include <stdio.h>
-        #ifdef NCCL_MAJOR
-        #ifndef NCCL_VERSION_CODE
-        #  define NCCL_VERSION_CODE \
-                (NCCL_MAJOR * 1000 + NCCL_MINOR * 100 + NCCL_PATCH)
-        #endif
-        #else
-        #  define NCCL_VERSION_CODE 0
-        #endif
-        int main() {
-          printf("%d", NCCL_VERSION_CODE);
-          return 0;
-        }
-        ''', include_dirs=settings['include_dirs'])
+        out = build_and_run(compiler,
+                            '''
+                            #ifndef CUPY_USE_HIP
+                            #include <nccl.h>
+                            #else
+                            #include <rccl.h>
+                            #endif
+                            #include <stdio.h>
+                            #ifdef NCCL_MAJOR
+                            #ifndef NCCL_VERSION_CODE
+                            #  define NCCL_VERSION_CODE \
+                            (NCCL_MAJOR * 1000 + NCCL_MINOR * 100 + NCCL_PATCH)
+                            #endif
+                            #else
+                            #  define NCCL_VERSION_CODE 0
+                            #endif
+                            int main() {
+                              printf("%d", NCCL_VERSION_CODE);
+                              return 0;
+                            }
+                            ''',
+                            include_dirs=settings['include_dirs'],
+                            define_macros=settings['define_macros'])
 
     except Exception as e:
         utils.print_warning('Cannot include NCCL\n{0}'.format(e))
@@ -570,6 +638,44 @@ def get_cub_version(formatted=False):
     return _cub_version
 
 
+def check_jitify_version(compiler, settings):
+    global _jitify_version
+
+    try:
+        cupy_jitify_include = _jitify_path
+        # Unfortunately Jitify does not have any identifiable name (branch,
+        # tag, etc), so we must use the commit here
+        a = subprocess.run(' '.join(['git', 'rev-parse', '--short', 'HEAD']),
+                           stdout=subprocess.PIPE, stderr=subprocess.PIPE,
+                           shell=True, cwd=cupy_jitify_include)
+        if a.returncode == 0:
+            out = a.stdout.decode()[:-1]  # unlike elsewhere, out is a str here
+        else:
+            raise RuntimeError('Cannot determine Jitify version from git')
+    except Exception as e:
+        utils.print_warning('Cannot determine Jitify version\n{}'.format(e))
+        # 0: Jitify is not built (makes no sense), -1: built with unknown ver
+        out = -1
+
+    _jitify_version = out
+    settings['define_macros'].append(('CUPY_JITIFY_VERSION_CODE',
+                                      _jitify_version))
+    return True  # we always build Jitify
+
+
+def get_jitify_version(formatted=False):
+    """Return Jitify version cached in check_jitify_version()."""
+    global _jitify_version
+    if _jitify_version is None:
+        msg = 'check_jitify_version() must be called first.'
+        raise RuntimeError(msg)
+    if formatted:
+        if _jitify_version == -1:
+            return '<unknown>'
+        return _jitify_version
+    raise RuntimeError('Jitify version is a commit string')
+
+
 def check_cutensor_version(compiler, settings):
     global _cutensor_version
     try:
@@ -614,6 +720,82 @@ def get_cutensor_version(formatted=False):
     return _cutensor_version
 
 
+def check_cusparselt_version(compiler, settings):
+    global _cusparselt_version
+    try:
+        out = build_and_run(compiler, '''
+        #include <cusparseLt.h>
+        #include <stdio.h>
+        #ifndef CUSPARSELT_VERSION
+        #define CUSPARSELT_VERSION 0
+        #endif
+        int main(int argc, char* argv[]) {
+          printf("%d", CUSPARSELT_VERSION);
+          return 0;
+        }
+        ''', include_dirs=settings['include_dirs'])
+
+    except Exception as e:
+        utils.print_warning('Cannot check cuSPARSELt version\n{0}'.format(e))
+        return False
+
+    _cusparselt_version = int(out)
+    return True
+
+
+def get_cusparselt_version(formatted=False):
+    """Return cuSPARSELt version cached in check_cusparselt_version()."""
+    global _cusparselt_version
+    if _cusparselt_version is None:
+        msg = 'check_cusparselt_version() must be called first.'
+        raise RuntimeError(msg)
+    return _cusparselt_version
+
+
+def check_cugraph_version(compiler, settings):
+    global _cugraph_version
+    try:
+        build_and_run(compiler, '''
+        #include <stdio.h>
+        #include <cugraph/raft/error.hpp>
+        int main(int argc, char* argv[]) {
+          return 0;
+        }
+        ''', include_dirs=settings['include_dirs'])
+    except Exception as e:
+        utils.print_warning('Cannot find cuGraph header files\n{0}'.format(e))
+        return False
+
+    try:
+        out = build_and_run(compiler, '''
+        #include <stdio.h>
+        #include <cugraph/version_config.hpp>
+        int main(int argc, char* argv[]) {
+          printf("%d", CUGRAPH_VERSION_MAJOR * 10000
+                     + CUGRAPH_VERSION_MINOR * 100
+                     + CUGRAPH_VERSION_PATCH);
+          return 0;
+        }
+        ''', include_dirs=settings['include_dirs'])
+    except Exception as e:
+        utils.print_warning('Cannot find cuGRAPH version information\n{0}'.
+                            format(e))
+        _cugraph_version = 0
+        return True
+
+    _cugraph_version = int(out)
+    return True
+
+
+def get_cugraph_version(formatted=False):
+    """Return cuGraph version cached in check_cugraph_version()."""
+    global _cugraph_version
+    if _cugraph_version is None:
+        msg = 'check_cugraph_version() must be called first.'
+        raise RuntimeError(msg)
+    return _cugraph_version
+
+
 def build_shlib(compiler, source, libraries=(),
                 include_dirs=(), library_dirs=(), define_macros=None,
                 extra_compile_args=()):
